/*
 * Lua Commander - Lua Console
 *
 * Copyright (C) 1999-2000, Antonio E. Scuri, Tecgraf/PUC-Rio.
 */


#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include <iup.h>
#include <iupkey.h>
#include <lua.h>
#include <luadebug.h>

#include "lconsole.h"
#include "utl.h"
#include "utl_dlist.h"


enum {DIR_DOWN, DIR_UP};

static struct _luaConsole
{
  Ihandle *iupDialog,
          *txCmdLine,
          *mlOutput;
  utlDList* command_list;           /* command history list */
  utlDListNode* cur_list_node;      /* contains the last command entered at the bottom of the list */
  int ind_level,                    /* table identation level */
      close,                        /* return value of the Close callback */
      output_size,                  /* size of the output buffer */
      command_dir,                  /* direction up/down in the history */
      hold_caret;                   /* flag that disables caret update, used for long prints */
  lua_Object print_table;           /* lua function that prints a table */
}console;

/* used by the print functions */
static void HoldCaret(int hold);

/* Lua isstring also consider numbers as strings */
int lua_rawisstring(lua_Object o) 
{
  static int string_tag = LUA_ANYTAG;

  if (string_tag == LUA_ANYTAG) 
  {  
    lua_pushstring("");
    string_tag = lua_tag(lua_pop());
  }

  return (lua_tag(o) == string_tag);
}

/*************************************************************************/
/*                      Print                                            */
/*************************************************************************/

/* the use of Next functions is done in Lua */

static char lua_list_func_define[] = {
"function lcListFunc()\n"
"  local n,v = nextvar(nil)\n"
"  lcHoldCaret(1)\n"
"  print()\n"
"  while n ~= nil do\n"
"    if type(v) == \"function\" then\n"
"      lcPrintFuncVar(n, v)\n"
"    end\n"
"    n,v = nextvar(n)\n"
"  end\n"
"  lcHoldCaret(0)\n"
"end\n"
};

static char lua_list_var_define[] = {
"function lcListVar()\n"
"  local n,v = nextvar(nil)\n"
"  lcHoldCaret(1)\n"
"  print()\n"
"  while n ~= nil do\n"
"    if type(v) ~= \"function\" then\n"
"      lcPrintVar(n, v)\n"
"    end\n"
"    n,v = nextvar(n)\n"
"  end\n"
"  lcHoldCaret(0)\n"
"end\n"
};

static char lua_print_table_define[] = {
"function lcPrintTable(t)\n"
"  local n,v = next(t, nil)\n"
"  while n ~= nil do\n"
"    lcPrintVar(n, v, 1)\n"
"    n,v = next(t, n)\n"
"  end\n"
"end\n"
};

static char* msg_indent(void)
{
  switch(console.ind_level)
  {
  case 0:
    return "";
  case 1:
    return "  ";
  case 2:
    return "    ";
  case 3:
    return "      ";
  case 4:
    return "        ";
  case 5:
    return "          ";
  case 6:
    return "            ";
  }
  return "";
}

static void lua_print_number(char* name, lua_Object o)
{
  double number = lua_getnumber(o);

  /* check if it an integer */
  if (number == floor(number))
    luaConsoleEnterMessagef("%s%s = %d", msg_indent(), name, (int)number);
  else
    luaConsoleEnterMessagef("%s%s = %f", msg_indent(), name, number);
}

static void lua_print_string(char* name, lua_Object o)
{
  luaConsoleEnterMessagef("%s%s = \"%s\"", msg_indent(), name, lua_getstring(o));
}

static void lua_print_userdata(char* name, lua_Object o)
{
  luaConsoleEnterMessagef("%s%s = <userdata> (tag = %d)", msg_indent(), name, lua_tag(o));
}

static void lua_print_table(char* name, lua_Object table)
{
  int hold_caret;

  if (console.ind_level == 6)
  {
    luaConsoleEnterMessagef("%s{table too deep}", msg_indent());
    return;
  }

  luaConsoleEnterMessagef("%s%s = ", msg_indent(), name);
  luaConsoleEnterMessagef("%s{", msg_indent());
  console.ind_level++;

  hold_caret = console.hold_caret;
  if (!hold_caret) HoldCaret(1);

  lua_pushobject(table);
  lua_callfunction(console.print_table);

  if (!hold_caret) HoldCaret(0);

  console.ind_level--;
  luaConsoleEnterMessagef("%s}", msg_indent());
}

static void lua_print_function(char* name, lua_Object o)
{
  char* filename, *obj_name, *type;
  int linedefined;

  lua_funcinfo(o, &filename, &linedefined);
  type = lua_getobjname(o, &obj_name);

  if (lua_iscfunction(o))
  {
    if (*type == 't')
      luaConsoleEnterMessagef("%s%s = <cfunction> (\"%s\" tag method)", msg_indent(), name, obj_name);
    else
      luaConsoleEnterMessagef("%s%s = <cfunction>", msg_indent(), name);
  }
  else if (lua_isfunction(o))
  {
    if (*type == 't')
    {
      if (filename[0] != '@')
        luaConsoleEnterMessagef("%s%s = <function> (\"%s\" tag method) [defined in a string at line %d]", msg_indent(), name, obj_name, linedefined);
      else
      {
        char* text = utlGetFileTitle(filename+1);
        luaConsoleEnterMessagef("%s%s = <function> (\"%s\" tag method) [defined in the file \"%s\" at line %d]", msg_indent(), name, obj_name, text, linedefined);
        free(text);
      }
    }
    else
    {
      if (filename[0] != '@')
        luaConsoleEnterMessagef("%s%s = <function> [defined in a string at line %d]", msg_indent(), name, linedefined);
      else
      {
        char* text = utlGetFileTitle(filename+1);
        luaConsoleEnterMessagef("%s%s = <function> [defined in the file \"%s\" at line %d]", msg_indent(), name, text, linedefined);
        free(text);
      }
    }
  }
}

void luaConsolePrintObject(char* name, lua_Object o)
{
  if (lua_rawisstring(o))
    lua_print_string(name, o);
  else if (lua_isnumber(o))
    lua_print_number(name, o);
  else if (lua_istable(o))
    lua_print_table(name, o);
  else if (lua_isuserdata(o))
    lua_print_userdata(name, o);
  else if (lua_isnil(o))
    luaConsoleEnterMessagef("%s%s = nil", msg_indent(), name);
  else if (lua_isfunction(o))
    lua_print_function(name, o);
  else
    luaConsoleEnterMessagef("%s%s = <unknown>", msg_indent(), name);
}

static void lua_print_func_var(void)
{
  /* used only by the lcListFunc */
  lua_Object name, func;
  name = lua_getparam(1);
  func = lua_getparam(2);

  if (name == LUA_NOOBJECT || func == LUA_NOOBJECT || !lua_isstring(name) || !lua_isfunction(func))
    lua_error("lcPrintFuncVar: invalid parameters.");

  if (lua_iscfunction(func))
    luaConsoleEnterMessagef("%s  (C)", lua_getstring(name));
  else
    luaConsoleEnterMessage(lua_getstring(name));
}

static void lua_print_var(void)
{
  /* used only by the lcListVar and lcPrintTable */
  lua_Object n, v, table_flag;
  char name[512];

  lua_beginblock();  
  
  n = lua_getparam(1);
  v = lua_getparam(2);
  table_flag = lua_getparam(3);

  if (table_flag == LUA_NOOBJECT)
    sprintf(name, "%s", lua_getstring(n));
  else
  {
    if (lua_rawisstring(n))
      sprintf(name, "[\"%s\"]", lua_getstring(n));
    else
      sprintf(name, "[%s]", lua_getstring(n));
  }

  luaConsolePrintObject(name, v);

  lua_endblock();  
}

static void lua_print_result(void)
{
  lua_Object o;
  char buf[50];
  int i = 1;

  o = lua_getresult(i);
  while (o != LUA_NOOBJECT)
  {
    lua_beginblock();  

    if (lua_getresult(2) == LUA_NOOBJECT)
      strcpy(buf, "");
    else
      sprintf(buf, "<%d return>", i);

    console.ind_level = 0;
    luaConsolePrintObject(buf, o);

    i++;
    o = lua_getresult(i);

    lua_endblock();  
  }
}

/*************************************************************************/
/*                      Lua Overriden Functions                          */
/*************************************************************************/


static void lua_print(void)
{
  lua_Object o;
  int i = 1;
  char buffer[10240] = "";

  o = lua_getparam(i);
  while (o != LUA_NOOBJECT)
  {
    lua_beginblock();  

    if (i > 1)
      strcat(buffer, "\t");

    if (lua_isstring(o))
      strcat(buffer, lua_getstring(o));
    else if (lua_isnil(o))
      strcat(buffer, "nil");
    else if (lua_istable(o))
      strcat(buffer, "<table>");
    else if (lua_iscfunction(o))
      strcat(buffer, "<cfunction>");
    else if (lua_isfunction(o))
      strcat(buffer, "<function>");
    else if (lua_isuserdata(o))
      strcat(buffer, "<userdata>");
    else
      strcat(buffer, "<unknown>");

    i++;
    o = lua_getparam(i);

    lua_endblock();  
  }

  luaConsoleEnterMessage(buffer);
}

static void lua_error_msg_func(void)
{
  char buf[200];
  lua_Object o = lua_getparam (1);
  sprintf(buf, "lua: %s\n", lua_isstring(o) ? lua_getstring(o) : "(no messsage)");
  luaConsoleEnterMessage(buf);
}


/*************************************************************************/
/*                      Command Line                                     */
/*************************************************************************/


static int ConsoleUpList(void)
{
  /* callback called when the user press the Up key */

  if (console.command_list->total == 0) /* empty list */
    return IUP_IGNORE;

  if (console.cur_list_node != console.command_list->top && console.command_dir == DIR_DOWN) 
  {                                      /* I was going down and decide to go up, so turn */
    console.command_dir = DIR_UP;
    console.cur_list_node = console.cur_list_node->previous;
  }

  if (console.command_dir == DIR_DOWN)   /* pass the top, does nothing */
    return IUP_IGNORE;

  /* write and up */
  IupStoreAttribute(console.txCmdLine, IUP_VALUE, (char*)console.cur_list_node->value);
  console.cur_list_node = console.cur_list_node->previous;

  if (!console.cur_list_node)
  {
    console.cur_list_node = console.command_list->top;
    console.command_dir = DIR_DOWN;       /* pass the top, prepare to go down */
  }
  else
    console.command_dir = DIR_UP;

  return IUP_IGNORE;
}

static int ConsoleDownList(void)
{
  /* callback called when the user press the Down key */

  if (console.command_list->total == 0) /* empty list */
    return IUP_IGNORE;

  if (console.cur_list_node != console.command_list->bottom && console.command_dir == DIR_UP) 
  {                                      /* I was going up and decide to go down, so turn */
    console.command_dir = DIR_DOWN;
    console.cur_list_node = console.cur_list_node->next;
  }

  if (console.command_dir == DIR_UP)     /* pass the bottom, does nothing */
    return IUP_IGNORE;

  /* down and write */
  console.cur_list_node = console.cur_list_node->next;
  if (console.cur_list_node == NULL)    /* pass the bottom, prepare to go up, and clean up */
  {
    console.cur_list_node = console.command_list->bottom;
    console.command_dir = DIR_UP;
    IupStoreAttribute(console.txCmdLine, IUP_VALUE, "");      
  }
  else
  {
    console.command_dir = DIR_DOWN;
    IupStoreAttribute(console.txCmdLine, IUP_VALUE, (char*)console.cur_list_node->value);
  }

  return IUP_IGNORE;
}

static void dummy_error_func(void)
{
  /* used only by ConsolePrintExpression */
}

static void ConsolePrintExpression(char *command)
{
  int err;
  lua_Object old_func;
  char *buffer = malloc(strlen(command)+10);

  /* used only by luaConsoleEnterCommand */

  /* remove the error function */
  lua_pushcfunction(dummy_error_func);
  old_func = lua_seterrormethod();

  /* compile an expression that can be evaluated. */
  sprintf(buffer, "return (%s)", command);
  err = lua_dostring(buffer);
  if (err==0)
  {
    /* no problems, just print the results */
    lua_print_result();

    /* Examples: */
      /* x, where x is nil */
      /* 3+5 */
      /* t[1] */
      /* sin(0) */

    /* restores the error function */
    lua_pushobject(old_func);
    lua_seterrormethod();
  }
  else
  {
    /* oops, can be an error or the expression can not be evaluated */

    /* restores the error function */
    lua_pushobject(old_func);
    lua_seterrormethod();

    /* just execute the string, let the error handler inform a possible error */
    lua_dostring(command);

    /* Examples: */
      /* t=3  */
      /* if x==nil then x = sin(0) else x = sin(10) end */
  }

  free(buffer);
}

void luaConsoleEnterCommand(char *command)
{
  int hold_caret;
  lua_Object o;

  lua_beginblock();  
  hold_caret = console.hold_caret;
  if (!hold_caret) HoldCaret(1);

  command = utlStrDup(command);

  /* prints the command in the console */
  luaConsoleEnterCommandStr(command);

  o = lua_getglobal(command);
  if (!lua_isnil(o))
  {
    console.ind_level = 0;
    luaConsolePrintObject(command, o);
  }
  else
  {
    /* if its not an object, may be an expresssion */
    ConsolePrintExpression(command);
  }

  free(command);

  if (!hold_caret) HoldCaret(0);
  lua_endblock();  

  IupSetAttribute(console.txCmdLine, IUP_VALUE, "");
  IupSetFocus(console.txCmdLine);
}


/*************************************************************************/
/*                   Console  Multiline                                  */
/*************************************************************************/


static void HoldCaret(int hold)
{
  if (!hold) IupSetAttribute(console.mlOutput, IUP_CARET, "65535,65535");
  console.hold_caret = hold;
}

static void luaHoldCaret(void)
{
  int hold;
  hold = (int)lua_getnumber(lua_getparam(1));
  HoldCaret(hold);
}

static int ConsoleClear(void)
{
  IupSetAttribute(console.txCmdLine, IUP_VALUE, "");
  return IUP_DEFAULT;
}

static int ConsoleEnterKey(void)
{
  char* text = IupGetAttribute(console.txCmdLine, IUP_VALUE);
  if (!text || text[0] == 0)
    return IUP_DEFAULT;

  luaConsoleEnterCommand(text);
  return IUP_DEFAULT;
}

int luaConsoleOutputSizeDlg(void)
{
  int size = luaConsoleOutputSize(0);
  if (IupScanf("Console Output\nSize: %5.10%d\n", &size) == -1)
    return IUP_DEFAULT;

  luaConsoleOutputSize(size);

  return IUP_DEFAULT;
}

int luaConsoleOutputSize(int size)
{
  int old_size = console.output_size;
  if (size == 0)
    return old_size;

  console.output_size = size;

  if (size < old_size)
  {
    char* value = IupGetAttribute(console.mlOutput, IUP_VALUE);

    if ((int)strlen(value) > size)
    {
      char* buffer;
      int len = size-1;
      while(len > 0 && value[len] != '\n')
        len--;

      buffer = utlStr(len+2);
      memcpy(buffer, value, len);
  
      buffer[len] = '\n';
      buffer[len+1] = 0;
  
      IupStoreAttribute(console.mlOutput, IUP_VALUE, buffer);
      if (!console.hold_caret) IupSetAttribute(console.mlOutput, IUP_CARET, "65535,65535");
      free(buffer);
    }
  }

  return old_size;
}

char* luaConsoleGetOutput(void)
{
  return IupGetAttribute(console.mlOutput, IUP_VALUE);
}

int luaConsoleClearOutput(void)
{
  IupSetAttribute(console.mlOutput, IUP_VALUE, "");
  return IUP_DEFAULT;
}

void luaConsoleEnterCommandStrf(char* format, ...)
{
  char value[65535];
  va_list arglist;
  va_start(arglist, format);
  vsprintf(value, format, arglist);

  luaConsoleEnterCommandStr(value);
}

void luaConsoleEnterCommandStr(char* text)
{
  utlDListInsertBottom(console.command_list, utlStrDup(text));
  console.cur_list_node = console.command_list->bottom;
  luaConsoleEnterMessagef("> %s", text);
}

void luaConsoleEnterMessagef(char* format, ...)
{
  char value[65535];
  va_list arglist;
  va_start(arglist, format);
  vsprintf(value, format, arglist);

  luaConsoleEnterMessage(value);
}

void luaConsoleEnterMessage(char* text)
{
  char *buffer;
  char* value = IupGetAttribute(console.mlOutput, IUP_VALUE);
  int start_value = 0, start_text = 0;

  int len_text = strlen(text);
  int len_value = strlen(value);

  if (len_value + len_text > console.output_size)
  {
    start_value = len_value + len_text - console.output_size;

    if (start_value > len_value)
    {
      len_value = 0;
      if (len_text > console.output_size)
      {
        start_text = len_text - console.output_size;
        len_text = console.output_size-2;
      }
    }
    else
    {
      while(start_value < len_value && value[start_value] != '\n')
        start_value++;

      start_value++;
      len_value -= start_value;
    }
  }

  buffer = utlStr(len_text+len_value+2);
  if (len_value) memcpy(buffer, value + start_value, len_value);
  memcpy(buffer+len_value, text+start_text, len_text);

  buffer[len_text+len_value] = '\n';
  buffer[len_text+len_value+1] = 0;
  
  IupStoreAttribute(console.mlOutput, IUP_VALUE, buffer);
  if (!console.hold_caret) IupSetAttribute(console.mlOutput, IUP_CARET, "65535,65535");
  free(buffer);
}

static void ConsoleEnterMessage(void)
{
  char* text;
  text = lua_getstring(lua_getparam(1));
  if (text) luaConsoleEnterMessage(text);
}

static void ConsoleEnterCommandStr(void)
{
  char* text;
  text = lua_getstring(lua_getparam(1));
  if (text) luaConsoleEnterCommandStr(text);
}

static void ConsoleEnterCommand(void)
{
  char* text;
  text = lua_getstring(lua_getparam(1));
  if (text) luaConsoleEnterCommand(text);
}


/*************************************************************************/
/*                      Window                                           */
/*************************************************************************/


int utlGetLuaFile(char* filename, int save)
{
  Ihandle *filedlg; 
  int result = 0; 
  char* file_name;

  filedlg = IupFileDlg(); 

  if (save)
  {
    IupSetAttribute(filedlg, IUP_DIALOGTYPE, IUP_SAVE); 
    IupSetAttribute(filedlg, IUP_ALLOWNEW, IUP_YES); 
    IupSetAttribute(filedlg, IUP_FILE, filename); 
  }
  else
  {
    IupSetAttribute(filedlg, IUP_DIALOGTYPE, IUP_OPEN); 
    IupSetAttribute(filedlg, IUP_ALLOWNEW, IUP_NO); 
  }

  IupStoreAttribute(filedlg, IUP_DIRECTORY, utlGetCurrentDirectory()); 
  IupSetAttribute(filedlg, IUP_TITLE, "Select File"); 
  IupSetAttribute(filedlg, IUP_FILTER, "*.lua"); 
  IupSetAttribute(filedlg, IUP_FILTERINFO, "Lua files (*.lua)"); 
  IupSetAttribute(filedlg, IUP_NOCHANGEDIR, IUP_NO); 

  IupPopup(filedlg, IUP_CENTER, IUP_CENTER); 

  result = IupGetInt(filedlg, IUP_VALUE);
  if (result == -1) return 0;

  file_name = IupGetAttribute(filedlg, IUP_VALUE);
  if (!file_name) return 0;

  strcpy(filename, file_name);

  return 1;
}

static int ConsoleRunFile(void)
{
  char* title;
  char FileName[1512] = "*.lua";

  if (!utlGetLuaFile(FileName, 0))
    return IUP_DEFAULT;

  title = utlGetFileTitle(FileName);
  luaConsoleEnterCommandStrf("dofile(\"%s\")", title);
  lua_dofile(FileName);
  free(title);
 
  return IUP_DEFAULT;
}

static void About(void)
{
  char buffer[512];
  sprintf(buffer, 
    "Lua Commander 1.0\n"
    "  (written by A. Scuri)\n"
    "%s  %s\n"
    "  (written by %s)\n" 
    "IUP %s\n"
    "  (copyright (c) 1995-2000 Tecgraf/PUC-Rio - SEPROC/CENPES/Petrobrs)\n", 
    LUA_VERSION, LUA_COPYRIGHT, LUA_AUTHORS, IupGetGlobal(IUP_VERSION));
  luaConsoleEnterMessage(buffer);
}

static int ConsoleOutput(void)
{
  Ihandle* menu;

  menu = IupMenu(
    IupItem ("Clear\tCtrl+Del", "luaConsoleClearOutput"),
    IupItem ("Buffer Size...", "luaConsoleOutputSizeDlg"),
    NULL);

  IupPopup(menu,IUP_MOUSEPOS,IUP_MOUSEPOS);

  return IUP_DEFAULT;
}

static int ConsoleListFunc(void)
{
  luaConsoleEnterCommandStr("lcListFunc()");
  lua_call("lcListFunc");
  return IUP_DEFAULT;
}

static int ConsoleListVar(void)
{
  luaConsoleEnterCommandStr("lcListVar()");
  lua_call("lcListVar");
  return IUP_DEFAULT;
}

static int ConsoleList(void)
{
  Ihandle* menu;

  menu = IupMenu(
    IupItem ("Functions", "luaConsoleListFunc"),
    IupItem ("Variables", "luaConsoleListVar"),
    NULL);

  IupPopup(menu,IUP_MOUSEPOS,IUP_MOUSEPOS);

  return IUP_DEFAULT;
}

int luaConsoleShow(void)
{
  IupShow(console.iupDialog);
  return IUP_DEFAULT;
}

static int ConsoleClose(void)
{
  return console.close;
}

void luaConsoleKill(void)
{
  utlKillDList(console.command_list, free);
  IupDestroy(console.iupDialog);
}

void luaConsoleCreate(char* parent, int close)
{
  Ihandle *vbLabels, *hbToolBar;

  memset(&console, 0, sizeof(console));

  console.output_size = 32768;
  console.close = close;
  console.command_list = utlCreateDList();
  console.cur_list_node = NULL;
  console.command_dir = DIR_UP;

  IupSetFunction("luaConsoleOutputSizeDlg", (Icallback)luaConsoleOutputSizeDlg);
  IupSetFunction("luaConsoleClearOutput", (Icallback)luaConsoleClearOutput);
  IupSetFunction("luaConsoleClose", (Icallback)ConsoleClose);
  IupSetFunction("luaConsoleEnterKey", (Icallback)ConsoleEnterKey);
  IupSetFunction("luaConsoleUpList", (Icallback)ConsoleUpList);
  IupSetFunction("luaConsoleDownList", (Icallback)ConsoleDownList);
  IupSetFunction("luaConsoleClear", (Icallback)ConsoleClear);
  IupSetFunction("luaConsoleRunFile", (Icallback)ConsoleRunFile);
  IupSetFunction("luaConsoleOutput", (Icallback)ConsoleOutput);
  IupSetFunction("luaConsoleList", (Icallback)ConsoleList);
  IupSetFunction("luaConsoleListFunc", (Icallback)ConsoleListFunc);
  IupSetFunction("luaConsoleListVar", (Icallback)ConsoleListVar);

  hbToolBar = IupHbox(
    IupSetAttributes(IupLabel(""),"IMAGE=LuaButton"),
    IupFill(),
    IupSetAttributes(IupButton("Do File...", "luaConsoleRunFile"), "TIP=\"Executes the LUA code in a file.\", SIZE=50"),
    IupSetAttributes(IupFill(), "EXPAND=NO, SIZE=15"),
    IupSetAttributes(IupButton("Output >>", "luaConsoleOutput"), "TIP=\"Output control.\""),
    IupSetAttributes(IupFill(), "EXPAND=NO, SIZE=3"),
    IupSetAttributes(IupButton("Global List >>", "luaConsoleList"), "TIP=\"Lists global functions and variables.\""),
    NULL);

  IupSetAttributes(hbToolBar,"MARGIN=0x0, GAP=0, ALIGNMENT=ACENTER");

  console.mlOutput = IupSetAttributes(IupMultiLine("do_nothing"), "EXPAND=YES");
  console.txCmdLine = IupSetAttributes(IupText("do_nothing"), "EXPAND=HORIZONTAL");

  vbLabels = IupVbox(
    hbToolBar,
    IupSetAttributes(IupFrame(IupVbox(console.txCmdLine, NULL)), "TITLE=\"Command Line:\""),
    IupSetAttributes(IupFrame(IupVbox(console.mlOutput, NULL)), "TITLE=Output:"),
    NULL);

  IupSetAttribute(console.txCmdLine,IUP_K_UP, "luaConsoleUpList");
  IupSetAttribute(console.txCmdLine,IUP_K_DOWN, "luaConsoleDownList");
  IupSetAttribute(console.txCmdLine,IUP_K_ESC, "luaConsoleClear");
  IupSetAttribute(console.txCmdLine,IUP_FONT, IUP_COURIER_NORMAL_12);
  IupSetAttribute(console.mlOutput,IUP_READONLY, IUP_YES);
  IupSetAttribute(console.mlOutput,IUP_BGCOLOR, "224 224 224");
  IupSetAttribute(console.mlOutput,IUP_FONT, IUP_COURIER_NORMAL_12);
  IupSetAttributes(vbLabels,"MARGIN=5x5, GAP=5");

  console.iupDialog = IupDialog(vbLabels);
  if (parent) IupSetAttribute(console.iupDialog,IUP_PARENTDIALOG, parent);
  IupSetAttribute(console.iupDialog,IUP_CLOSE_CB,"luaConsoleClose");
  IupSetAttribute(console.iupDialog,IUP_TITLE, "Lua Console");
  IupSetAttribute(console.iupDialog,IUP_K_CR, "luaConsoleEnterKey");
  IupSetAttribute(console.iupDialog,IUP_K_cDEL, "luaConsoleClearOutput");

  IupSetHandle("luaConsole", console.iupDialog);
  IupSetAttribute(console.iupDialog,IUP_SIZE, "300x200");
  IupSetAttribute(console.iupDialog,IUP_ICON, "LuaIcon");

  IupShowXY(console.iupDialog, 50,50);

  lua_pushcfunction(lua_error_msg_func);
  lua_setglobal("_ERRORMESSAGE");
  lua_pushcfunction(lua_error_msg_func);
  lua_setglobal("_ALERT");

  lua_register("print", lua_print);

  lua_register("lcShow", (lua_CFunction)luaConsoleShow);
  lua_register("lcClearOutput", (lua_CFunction)luaConsoleClearOutput);
  lua_register("lcEnterMessage", ConsoleEnterMessage);
  lua_register("lcEnterCommandStr", ConsoleEnterCommandStr);
  lua_register("lcEnterCommand", ConsoleEnterCommand);
  lua_register("lcHoldCaret", luaHoldCaret);
  lua_register("lcAbout", About);

  lua_register("lcPrintFuncVar", lua_print_func_var);
  lua_dostring(lua_list_func_define);

  lua_register("lcPrintVar", lua_print_var);
  lua_dostring(lua_list_var_define);
  lua_dostring(lua_print_table_define);

  console.print_table = lua_getglobal("lcPrintTable");

  luaConsoleEnterCommandStr("lcAbout()");
  About();
}

